# -*- coding: utf-8 -*-
from __future__ import absolute_import

import dask
import yaml
import os
from .coffea_casa import CoffeaCasaCluster
from .plugin import DistributedEnvironmentPlugin
from . import config

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


def is_number(in_value):
    try:
        float(in_value)
        return True
    except ValueError:
        return False



"""
# coffea-casa/config.py
# https://docs.dask.org/en/latest/configuration.html
"""
import os
import yaml

import dask.config

fn = os.path.join(os.path.dirname(__file__), "jobqueue-coffea-casa.yaml")

with open(fn) as f:
    defaults = yaml.safe_load(f)
dask.config.update_defaults(defaults)

dask.config.ensure_file(source=fn, comment=True)



import os
import sys
import zipfile
import logging
import uuid
import subprocess

from distributed.diagnostics.plugin import NannyPlugin
from dask.utils import tmpfile

logger = logging.getLogger(__name__)


class DistributedEnvironmentPlugin(NannyPlugin):
    """A NannyPlugin to upload a local pip installable package to workers.
    Parameters
    ----------
    path: str
        A path to the pip installable directory to upload
    Examples
    --------
    >>> from distributed.diagnostics.plugin import DistributedEnvironmentPlugin
    >>> client.register_worker_plugin(DistributedEnvironmentPlugin("/path/to/directory"), nanny=True)  # doctest: +SKIP
    """

    def __init__(
        self,
        path,
        pip_options=None,
        restart=False,
        update_path=False,
        skip_words=(".git", ".github", ".pytest_cache", "tests", "docs"),
        skip=(lambda fn: os.path.splitext(fn)[1] == ".pyc",),
        extra_inputs=[],
    ):
        """
        Initialize the plugin by reading in the data from the given file.
        """
        path = os.path.expanduser(path)
        self.package = os.path.split(path)[-1]
        self.restart = restart
        self.update_path = update_path
        if pip_options is None:
            pip_options = []
        self.pip_options = pip_options
        self.extra_inputs = [os.path.split(x)[-1] for x in extra_inputs]

        self.name = "upload-directory-" + self.package

        with tmpfile(extension="zip") as fn:
            with zipfile.ZipFile(fn, "w", zipfile.ZIP_DEFLATED) as z:
                for root, dirs, files in os.walk(path):
                    for file in files:
                        filename = os.path.join(root, file)
                        if any(predicate(filename) for predicate in skip):
                            continue
                        dirs = filename.split(os.sep)
                        if any(word in dirs for word in skip_words):
                            continue

                        archive_name = os.path.relpath(
                            os.path.join(root, file), os.path.join(path, "..")
                        )
                        z.write(filename, archive_name)
                for fpath in extra_inputs:
                    fname = os.path.split(fpath)[-1]
                    z.write(fpath,fname)

            with open(fn, "rb") as f:
                self.data = f.read()

    def setup(self, nanny):
        # Copy the package to the worker machine
        logger.info("Entering plugin setup")
        logger.info("nanny.local_directory: %s",nanny.local_directory)
        logger.info("nanny.worker_dir: %s",nanny.worker_dir)
        fn = os.path.join(nanny.local_directory, f"tmp-{str(uuid.uuid4())}.zip")
        with open(fn, "wb") as f:
            f.write(self.data)

        with zipfile.ZipFile(fn) as z:
            z.extractall(path=nanny.local_directory)

        if self.update_path:
            if nanny.local_directory not in sys.path:
                sys.path.insert(0, nanny.local_directory)
            path = os.path.join(nanny.local_directory, self.package)
            if path not in sys.path:
                sys.path.insert(0, path)

        # Now try to pip install the package
        package_path = os.path.join(nanny.local_directory,self.package)
        logger.info("Installing the package: %s",self.package)
        proc = subprocess.Popen(
            [sys.executable, "-m", "pip", "install"] + self.pip_options + [package_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = proc.communicate()
        returncode = proc.wait()

        if returncode:
            logger.error("Pip install failed with '%s'",stderr.decode().strip())
            return

        # Cleanup the zip file
        logger.info("Cleaning up temporary directory: %s",fn)
        os.remove(fn)

        return

    def teardown(self, nanny):
        for fname in self.extra_inputs:
            logger.info(f"Removing: {fname}")
            path = os.path.join(nanny.local_directory,fname)
            os.remove(path)
        logger.info("Uninstalling the package: %s",self.package)
        proc = subprocess.Popen(
            [sys.executable, "-m", "pip", "uninstall", self.package],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = proc.communicate()
        returncode = proc.wait()

        if returncode:
            logger.error("Pip uninstall failed with '%s'",stderr.decode().strip())
            return

        return



"""Setup.py for coffea_casa
"""
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages
import io
import versioneer

# Add 'distributed' back when patches will be upstreamed
INSTALL_REQUIRES = ['dask-jobqueue']
TESTS_REQUIRE = ['pytest', 'pytest-cov'
                 'pytest-timeout', 'pytest-rerunfailures']
EXTRAS_REQUIRE = {
    'docs': ['ipython', 'sphinx', 'sphinx_rtd_theme',
             'sphinx-gallery', 'nbsphinx', 'nbstripout', 'docutils', 'sphinx_rtd_theme' , "mkdocs==1.2.4"],
    'test': ['pytest', 'pytest-cov', 'pytest-timeout', 'pytest-rerunfailures']
}

EXTRAS_REQUIRE['all'] = sorted(set(sum(EXTRAS_REQUIRE.values(), [])))

with io.open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(name='coffea_casa',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      packages=find_packages(),
      description='A Prototype U.S. CMS analysis facility',
      long_description=readme,
      url='http://github.com/CoffeaTeam/coffea-casa.git',
      author='UNL',
      author_email='oksana.shadura@cern.ch',
      license='BSD-3-Clause License',
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      include_package_data=True,
      zip_safe=False,
      long_description_content_type = 'text/markdown',
      setup_requires=["pytest-runner", "flake8"],
      dependency_links=['git+git://github.com/oshadura/distributed.git@coffea-casa-facility#egg=distributed'],
      tests_require=TESTS_REQUIRE,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
      ],
      platforms="Any",
      )



"""CoffeaCasaCluster class
"""
import os
import sys
from pathlib import Path
import dask
from dask_jobqueue.htcondor import HTCondorCluster, HTCondorJob
from distributed.security import Security

# Port settings
DEFAULT_SCHEDULER_PORT = 8786
DEFAULT_DASHBOARD_PORT = 8787
DEFAULT_CONTAINER_PORT = 8786
#DEFAULT_NANNY_PORT = 8001

# Security settings for Dask scheduler
# REMOVE ME (backward compatibity for now)
SECRETS_DIR = Path("/etc/cmsaf-secrets")
SECRETS_DIR_CHOWN = Path("/etc/cmsaf-secrets-chown")
# CEPH (Skyhook)
CEPH_DIR = Path("/opt/ceph")
CEPH_CONF = CEPH_DIR / "ceph.conf"
KEYRING_CONF = CEPH_DIR / "keyring"
CA_FILE = SECRETS_DIR / "ca.pem"
CERT_FILE = SECRETS_DIR / "hostcert.pem"
HOME_DIR = Path.home()
# XCache
# REMOVE ME (backward compatibity for now)
XCACHE_FILE = SECRETS_DIR / "xcache_token"
XCACHE_SCITOKEN_FILE = SECRETS_DIR_CHOWN / "access_token"
# pip
PIP_REQUIREMENTS = HOME_DIR / "requirements.txt"
# conda, with yml/yaml both supported
if (HOME_DIR / "environment.yaml").is_file():
    CONDA_ENV = HOME_DIR / "environment.yaml"
else:
    CONDA_ENV = HOME_DIR / "environment.yml"


def merge_dicts(*dict_args):
    """
    Given any number of dictionaries, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dictionaries.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


class CoffeaCasaJob(HTCondorJob):
    """

    """
    submit_command = "condor_submit -spool"
    config_name = "coffea-casa"


class CoffeaCasaCluster(HTCondorCluster):
    """
    This is a  subclass expanding settings for launch Dask via HTCondorCluster
    over HTCondor in US.CMS facility.
    """
    job_cls = CoffeaCasaJob
    config_name = "coffea-casa"

    def __init__(self,
                 *,
                 security=None,
                 worker_image=None,
                 scheduler_options=None,
                 scheduler_port=DEFAULT_SCHEDULER_PORT,
                 dashboard_port=DEFAULT_DASHBOARD_PORT,
                 #nanny_port=DEFAULT_NANNY_PORT,
                 **job_kwargs):
        """
        Parameters
        ----------
        worker_image:
            Defaults to ``coffeateam/coffea-casa-analysis``
            (https://hub.docker.com/r/coffeateam/coffea-casa-analysis).
            Check the default version of container in `~/.config/dask/jobqueue-coffea-casa.yaml` or
            `/etc/dask/jobqueue-coffea-casa.yaml`
        scheduler_port:
            Defaults to 8786.
        dashboard_port:
            Defaults to 8787.
        container_port:
            Defaults to 8786.
        nanny_port:
            Defaults to 8001.
        disk:
            Total amount of disk per job (defaults to 5 GiB).
        cores:
            Total number of cores per job (defaults to 4 cores).
        memory:
            Total amount of memory per job (defaults to 6 GB).
        scheduler_options : dict
            Extra scheduler options for the job (Dask specific).
        job_extra : dict
            Extra submit file attributes for the job (HTCondor specific).

        Examples
        --------
        >>> from coffea_casa import CoffeaCasaCluster
        >>> cluster = CoffeaCasaCluster()
        >>> cluster.scale(jobs=10)  # ask for 10 jobs
        >>> from dask.distributed import Client
        >>> client = Client(cluster)
        This also works with adaptive clusters and launches and kill workers based on load.
        >>> cluster.adapt(maximum_jobs=20)
        """
        if security:
            self.security = security
        job_kwargs = self._modify_job_kwargs(
            job_kwargs,
            security=security,
            worker_image=worker_image,
            scheduler_port=scheduler_port,
            dashboard_port=dashboard_port,
            #nanny_port=nanny_port,
        )
        # Instantiate args and parameters from parent abstract class security=security
        super().__init__(**job_kwargs)

    @classmethod
    def _modify_job_kwargs(cls,
                           job_kwargs,
                           *,
                           security=None,
                           worker_image=None,
                           scheduler_options=None,
                           scheduler_port=DEFAULT_SCHEDULER_PORT,
                           dashboard_port=DEFAULT_DASHBOARD_PORT,
                           #nanny_port=DEFAULT_NANNY_PORT
                           ):
        job_config = job_kwargs.copy()
        input_files = []
        if CEPH_CONF.is_file() and KEYRING_CONF.is_file():
            input_files += [CEPH_CONF, KEYRING_CONF]
        if PIP_REQUIREMENTS.is_file():
            input_files += [PIP_REQUIREMENTS]
        if CONDA_ENV.is_file():
            input_files += [CONDA_ENV]
        # If we have certs in env, lets try to use TLS
        if (CA_FILE.is_file() and CERT_FILE.is_file() and cls.security().get_connection_args("scheduler")["require_encryption"]):
            job_config["protocol"] = "tls://"
            job_config["security"] = cls.security()
            input_files += [CA_FILE, CERT_FILE, XCACHE_FILE, XCACHE_SCITOKEN_FILE]
        else:
            raise KeyError("Please check with system administarator why you do not have a certificate.")
        files = ", ".join(str(path) for path in input_files)
        ## Networking settings
        try:
            external_ip = os.environ["HOST_IP"]
        except KeyError:
            print(
                "Please check with system administarator why external IP was not assigned for you."
            )
            sys.exit(1)
        scheduler_protocol = job_config["protocol"]
        if external_ip:
            address_list = [external_ip, DEFAULT_SCHEDULER_PORT]
            external_address_short = ":".join(str(item) for item in address_list)
            full_address_list = [scheduler_protocol, external_address_short]
            external_address = "".join(str(item) for item in full_address_list)
            external_ip_string = '"' + external_address + '"'
        # HTCondor logging
        job_config["log_directory"] = "logs"
        job_config["silence_logs"] = "DEBUG"
        ## Scheduler settings
        # we need to pass and check protocol for scheduler
        # (format should be not 'tls://'' but 'tls')
        job_config["scheduler_options"] = merge_dicts(
            {
                "port": scheduler_port,
                "dashboard_address": str(dashboard_port),
                "protocol": scheduler_protocol.replace("://", ""),
                "external_address": external_address,
            },
            job_kwargs.get(
                "scheduler_options",
                dask.config.get(f"jobqueue.{cls.config_name}.scheduler-options"),
            ),
        )
        ## Job extra settings (HTCondor ClassAd)
        job_config["job_extra"] = merge_dicts(
            {
                "universe": "docker",
                "docker_image": worker_image or dask.config.get(f"jobqueue.{cls.config_name}.worker-image")
            },
            {
                #"container_service_names": "dask,nanny",
                #"dask_container_port": DEFAULT_CONTAINER_PORT,
                #"nanny_container_port": nanny_port,
                "container_service_names": "dask",
                "dask_container_port": DEFAULT_CONTAINER_PORT,
            },
            {"transfer_input_files": files},
            {"encrypt_input_files": files},
            {"transfer_output_files": ""},
            {"when_to_transfer_output": "ON_EXIT"},
            {"should_transfer_files": "YES"},
            {"Stream_Output": "False"},
            {"Stream_Error": "False"},
            {"+DaskSchedulerAddress": external_ip_string},
            job_kwargs.get(
                "job_extra", dask.config.get(f"jobqueue.{cls.config_name}.job-extra")
            ),
        )
        print(job_config)
        return job_config

    @classmethod
    def security(cls):
        """
        Return the Dask ``Security`` object used by CoffeaCasa.
        """
        ca_file = str(CA_FILE)
        cert_file = str(CERT_FILE)
        return Security(
            tls_ca_file=ca_file,
            tls_worker_cert=cert_file,
            tls_worker_key=cert_file,
            tls_client_cert=cert_file,
            tls_client_key=cert_file,
            tls_scheduler_cert=cert_file,
            tls_scheduler_key=cert_file,
            require_encryption=True,
        )
